//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/View/Data/DataAccessWidget.h
//! @brief     Defines class DataAccessWidget
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2022
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#ifndef BORNAGAIN_GUI_VIEW_DATA_DATAACCESSWIDGET_H
#define BORNAGAIN_GUI_VIEW_DATA_DATAACCESSWIDGET_H

#include "GUI/Model/Device/RealItem.h"
#include "GUI/Model/Job/JobItem.h"
#include <QWidget>

//! The DataAccessWidget class is a base for widgets, working data, stored in JobItem or RealItem.
//! DataAccessWidget provides an access to SpeculatDataItems and Data2DItems separately and
//! as a list of items that is required for simultaneous, synchronous work while plotting and
//! changing their properties.

class DataAccessWidget : public QWidget {
    Q_OBJECT
public:
    explicit DataAccessWidget(QWidget* parent = nullptr);
    ~DataAccessWidget() override;

    virtual void setJobOrRealItem(QObject* item);

    virtual void setFourierItem(Data2DItem* fftItem);

    virtual QList<QAction*> actionList();

    JobItem* jobItem() const;
    RealItem* realItem() const;
    QObject* jobRealBase() const;

    //... Access to concrete items:

    // Specular

    //! Real data as Data1DItem
    Data1DItem* realData1DItem() const;

    //! Simulated data as Data1DItem
    Data1DItem* simuData1DItem() const;

    //! Difference data as Data1DItem
    Data1DItem* diffData1DItem() const;

    // Intensity

    //! Real as Data2DItem
    Data2DItem* realData2DItem() const;

    //! Simulated as Data2DItem
    Data2DItem* simuData2DItem() const;

    //! Difference as Data2DItem
    Data2DItem* diffData2DItem() const;

    //... Access to lists of items:

    // Specular

    //! Simulated + real specular data as list, non-nullptr items
    QList<Data1DItem*> mainData1DItems() const;

    //! Simulated + real + difference specular data as list, non-nullptr items
    QList<Data1DItem*> allData1DItems() const;

    //! first element of allData1DItems() with nullptr check
    Data1DItem* currentData1DItem() const;

    // Intensity

    //! Simulated + real intensity data as list, non-nullptr items
    QList<Data2DItem*> mainData2DItems() const;

    //! Simulated + real + difference intensity data as list, non-nullptr items
    QList<Data2DItem*> allData2DItems() const;

    //! first element of allData2DItems() with nullptr check
    Data2DItem* currentData2DItem() const;

private:
    template <class T>
    T* realDataItem() const;

    template <class T>
    T* simuDataItem() const;

    template <class T>
    T* diffDataItem() const;

    // simuDataItem() + realItem()
    template <class T>
    QList<T*> mainDataItems() const;

    // diffDataItem() as QList
    template <class T>
    QList<T*> diffDataItems() const;

    // mainDataItems() + diffDataItems()
    template <class T>
    QList<T*> allDataItems() const;

    QObject* m_item;
    Data2DItem* m_fftItem;
};

template <class T>
T* DataAccessWidget::realDataItem() const
{
    if (jobItem()) {
        RealItem* real_item_from_job = jobItem()->realItem();
        if (real_item_from_job)
            return dynamic_cast<T*>(real_item_from_job->dataItem());
        else
            return nullptr;
    }
    if (realItem())
        return dynamic_cast<T*>(realItem()->dataItem());

    return nullptr;
}

template <class T>
T* DataAccessWidget::simuDataItem() const
{
    if (!jobItem())
        return nullptr;
    return dynamic_cast<T*>(jobItem()->simulatedDataItem());
}

template <class T>
T* DataAccessWidget::diffDataItem() const
{
    if (!jobItem())
        return nullptr;
    return dynamic_cast<T*>(jobItem()->diffDataItem());
}

template <class T>
QList<T*> DataAccessWidget::mainDataItems() const
{
    QList<T*> output;
    auto* simu = simuDataItem<T>();
    auto* real = realDataItem<T>();
    if (simu)
        output.append(simu);
    if (real)
        output.append(real);
    return output;
}

template <class T>
QList<T*> DataAccessWidget::diffDataItems() const
{
    QList<T*> output;
    auto* diff = diffDataItem<T>();
    if (diff)
        output.append(diff);
    return output;
}

template <class T>
QList<T*> DataAccessWidget::allDataItems() const
{
    return mainDataItems<T>() + diffDataItems<T>();
}

#endif // BORNAGAIN_GUI_VIEW_DATA_DATAACCESSWIDGET_H
